#include <vector>
#include <stdlib.h> /* rand() */
#include <time.h>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 240;

extern "C" unsigned char RAM[0x800];

static unsigned PrevMarioX = 0, NFrames=0;
static unsigned MarioX     = 0;
static int MarioMovement;
static int Pit=0;
static BisqBotVar Entry, Damage;

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    if(PC == 0x8082)
    {
        unsigned MarioXPos = RAM[0x86];
        unsigned MarioYPos = RAM[0xCE];
        unsigned ScrollBegin = RAM[0x73F];
        
        if(MarioYPos > 0xC0) Pit=1;
        unsigned MarioScreenX = (unsigned char)(MarioXPos-ScrollBegin+256);
        
        MarioX = MarioXPos;
        if(NFrames++ > 0)
        {
            //if(!Entry.Count())
            MarioMovement += (signed char)(MarioX - PrevMarioX);
        }
        PrevMarioX = MarioX;
    }
    
    if(PC == 0xB04C) // magic_jT9+2
    {
        //fprintf(stderr, "JT9 with A=%02X\n", A);
        switch(A)
        {
            case 1: // auto climb
            case 2: // enter horiz pipe
            case 3: // enter vert pipe
            case 4: // slide flagpole down
            case 5: // victory 
            case 9: // eat mushroom
            case 12: // eat flower
                Entry.Set(A);
                break;
            case 10:// eat damage
            case 11:// eat death
                Damage.Set(A);
                Pit=1;
                break;
            case 0: // regen level
            case 7: // start level
                break;
            case 8: // normal level activity
                break;
        }
    }
    if(PC == 0xCFEC)
    {
        Entry.Set(15); // Killed koopa
    }
}

static unsigned JumpRemain = 0;
static unsigned NextChange = 0;
static unsigned NextJump   = 0;

void BisqBotReset()
{
    PrevMarioX = 0;
    NFrames = 0;
    MarioX = 0;
    MarioMovement = 0;
    
    Entry.Clear();
    Damage.Clear();
    
    NextJump   = rand()%30;
    JumpRemain = 0;
    Pit=0;
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    if(Entry.Count()) return 0;
    
    static const unsigned char InputTable[] =
    {
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_L,
        K_R | K_L,
        0
    };
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = K_B;
    unsigned char Input = 0x00;
    
    if(!NextChange) 
    {
        LastInput = InputTable[rand()%sizeof(InputTable)];
        NextChange = rand()%10;
    }
    else --NextChange;
    
    if(B_State)
        { if(rand()%100 < 1) B_State=0; }
    else
        { if(rand()%100 < 95) B_State=K_B; }

    /*
    if(!(rand()%20))
    {
        Input |= K_D;
        if((rand()%5)) Input &= ~(K_R | K_L);
    }
    */
    
    if(!JumpRemain)
    {
        if(!NextJump)
        {
            JumpRemain = 1+rand()%60;
            NextJump   = rand()%200 + 1;
        } else --NextJump;
    }
    
    Input |= LastInput;
    if(JumpRemain) Input |= K_A;;
    Input |= B_State;
    
    return Input;
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    //fprintf(stderr, "Joypad=%02X\n", RAM[0x6FC]);
    
    if(Pit) return 0; // dead
    
    unsigned score1 = 0;
    for(unsigned a=0; a<6; ++a)
        score1 = score1*10 + RAM[0x7DD+a];
    //fprintf(stderr, "score1: %u\n", score1);
    
    unsigned movescore = MarioMovement + 512;
    unsigned goalscore = Entry.Count() * 2048;
    return (movescore + goalscore) * 1000000ULL + score1;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
    if(Pit) return BisqBotRestart;
    
    if(FrameNo > MAX_FRAMES) return BisqBotRestart;
    //fprintf(stderr, "%u: %d\n", FrameNo, MarioMovement);
    return BisqBotContinue;
}
